!
! *
! *  Elmer, A Finite Element Software for Multiphysical Problems
! *
! *  Copyright 1st April 1995 - , CSC - IT Center for Science Ltd., Finland
! * 
! *  This program is free software; you can redistribute it and/or
! *  modify it under the terms of the GNU General Public License
! *  as published by the Free Software Foundation; either version 2
! *  of the License, or (at your option) any later version.
! * 
! *  This program is distributed in the hope that it will be useful,
! *  but WITHOUT ANY WARRANTY; without even the implied warranty of
! *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! *  GNU General Public License for more details.
! *
! *  You should have received a copy of the GNU General Public License
! *  along with this program (in file fem/GPL-2); if not, write to the 
! *  Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, 
! *  Boston, MA 02110-1301, USA.
! *
! *****************************************************************************/
!
!

MODULE HashTable

  USE Lists

  IMPLICIT NONE

  TYPE HashValue_t
     CHARACTER(LEN=MAX_NAME_LEN) :: Block,Type
  END TYPE HashValue_t


  TYPE HashEntry_t
     TYPE(HashEntry_t), POINTER  :: Next
     CHARACTER(LEN=MAX_NAME_LEN) :: Key
     TYPE(HashValue_t), POINTER  :: Value
  END TYPE HashEntry_t


  TYPE HashBucket_t
     TYPE(HashEntry_t), POINTER :: Head
  END TYPE HashBucket_t


  TYPE HashTable_t
     INTEGER :: BucketSize, CurrentBucket, TotalEntries
     INTEGER :: MaxAvgEntries
     TYPE(HashBucket_t), POINTER :: Bucket(:)
     TYPE(HashEntry_t),  POINTER :: CurrentEntry
  END TYPE HashTable_t


CONTAINS

!
  FUNCTION HashCreate( InitialBucketSize, MaxAvgEntries ) RESULT(hash)
    TYPE(HashTable_t), POINTER :: Hash
    INTEGER :: InitialBucketSize, MaxAvgEntries

    INTEGER :: i, RoundBits, Stat

    NULLIFY( Hash )
    IF ( InitialBucketSize <= 0 ) THEN
       WRITE( Message, * )  'HashCreate: invalid initial size given: ',InitialBucketSize
       CALL Error( 'HashCreate', Message )
       RETURN
    END IF

!   
    RoundBits = CEILING( LOG(1.0d0*InitialBucketSize) / LOG(2.0d0) )
 
!   
    ALLOCATE( Hash )

    Hash % BucketSize = 2**RoundBits;

!   
    ALLOCATE( Hash % Bucket( Hash % BucketSize ), STAT=stat )

    IF ( stat /= 0 ) THEN
       CALL Error( 'HashCreate', &
               'Hash table initialize error: unable to allocate bucket.' )
       DEALLOCATE( Hash )
       NULLIFY( Hash )

       RETURN
    END IF

    DO i=1,Hash % BucketSize
       NULLIFY( Hash % Bucket(i) % Head )
    END DO

    Hash % TotalEntries  = 0
    Hash % MaxAvgEntries = MaxAvgEntries;

!   
!   if ( EqualKeys )
!     hash->EqualKeys = EqualKeys;
!   else
!     hash->EqualKeys = (int (*)(void *,void *))HashEqualKeys;
!
!   
!   if ( HashFunc )
!     hash->HashFunc = HashFunc;
!   else
!     hash->HashFunc = (int (*)(void *,int))HashStringFunc;
   END FUNCTION HashCreate

!
   FUNCTION HashStringFunc( key, size ) RESULT(Ind)
     CHARACTER(LEN=*) :: key
     INTEGER :: Size, Ind

     INTEGER :: i,keylen

     DO keylen=LEN(key),1,-1
       IF ( key(keylen:keylen) /= ' ' ) EXIT
     END DO

     Ind = 0
     DO i=1,keylen
        Ind = Ind*8 + ICHAR(key(i:i))
     END DO

     Ind = IAND( Ind, size-1 ) + 1
   END FUNCTION HashStringFunc

!
   FUNCTION HashEqualKeys( key1,key2 ) RESULT(equal)

     CHARACTER(LEN=*) :: key1,key2
     LOGICAL :: equal
     INTEGER :: n1,n2

     equal = .FALSE.

     n1 = LEN_TRIM(key1)
     n2 = LEN_TRIM(key2)
     IF ( n1 /= n2 ) RETURN

     equal = key1(1:n1) == key2(1:n1)
   END FUNCTION HashEqualKeys

!
   FUNCTION HashFind( hash, key, n ) RESULT( entry )

     TYPE(HashTable_t), POINTER :: Hash
     TYPE(HashEntry_t), POINTER :: Entry, ptr
     INTEGER :: n
     CHARACTER(LEN=*) :: key

!    
     n = HashStringFunc( key, Hash % BucketSize )

!    
    ptr => Hash % Bucket(n) % Head
    DO WHILE( ASSOCIATED(ptr) )
       IF ( HashEqualKeys( key, ptr % key ) ) EXIT
       ptr => ptr % Next
    END DO
    Entry => ptr
  END FUNCTION HashFind

!
  RECURSIVE FUNCTION HashAdd( hash, key, value ) RESULT(Success)
    TYPE(HashTable_t), POINTER :: hash
    CHARACTER(LEN=*) :: key
    LOGICAL :: Success
    TYPE(HashValue_t), POINTER :: Value

    INTEGER :: stat
    INTEGER :: n, keylen, count = 0
    TYPE(HashEntry_t), POINTER :: Entry

    Success = .TRUE.

    entry => HashFind( hash,key, n )

    IF ( ASSOCIATED( entry ) ) THEN
!      
       Entry % Value => Value
    ELSE
!       
       ALLOCATE( Entry, STAT=stat )

       IF ( stat /= 0 ) THEN
          CALL Error( 'HashAdd', 'add failed: unable to allocate ' // &
               '(a few bytes of) memory ?' )
          RETURN
       END IF

       Entry % Next  => Hash % Bucket(n) % Head
       Entry % Value => Value
       Entry % Key = ' '
       DO keylen=LEN(key),1,-1
          IF ( key(keylen:keylen) /= ' ' ) EXIT
       END DO
       Entry % Key(1:keylen) = key(1:keylen)

       Hash % Bucket(n) % Head => Entry
       Hash % TotalEntries = Hash % TotalEntries + 1
       
       IF ( Hash % TotalEntries > Hash % MaxAvgEntries*Hash % BucketSize ) THEN
          Success = HashRebuild( Hash )
       END IF
    END IF
  END FUNCTION HashAdd

!
  SUBROUTINE HashRemove( Hash, key )
    TYPE(HashTable_t), POINTER :: Hash
    CHARACTER(LEN=*) :: Key

    TYPE(HashEntry_t), POINTER :: entry,prev
    INTEGER :: k,n

    IF ( .NOT. ASSOCIATED(hash) ) RETURN

!    
    n = HashStringFunc( key, hash % BucketSize );

!    
    NULLIFY( Prev )
    Entry => Hash % Bucket(n) % Head
    DO WHILE( ASSOCIATED( Entry ) )
!       
       IF ( HashEqualKeys( key,entry % key ) ) THEN
          IF ( ASSOCIATED(prev) ) THEN
            prev % next => entry % next
          ELSE
            hash % Bucket(n) % Head => Entry % next
         END IF

         DEALLOCATE(Entry)

         hash  % TotalEntries = Hash % TotalEntries - 1
         RETURN
      END IF
      Prev  => Entry
      Entry => Entry % Next
   END DO
 END SUBROUTINE HashRemove

!
 SUBROUTINE HashClean( hash )
   TYPE(HashTable_t), POINTER :: Hash

   TYPE(HashEntry_t), POINTER :: ptr,ptr1
   INTEGER :: i

   IF ( .NOT.ASSOCIATED(hash) ) RETURN

   DO i=1,hash % BucketSize
      ptr => hash % Bucket(i) % Head
      DO WHILE( ASSOCIATED(ptr) )
         ptr1 => ptr % next
         DEALLOCATE( ptr )
         ptr => ptr1
      END DO
      NULLIFY( hash % Bucket(i) % Head )
   END DO
   hash % TotalEntries  =  0
   hash % CurrentBucket =  0
   NULLIFY( hash % CurrentEntry )
 END SUBROUTINE HashClean

!
 SUBROUTINE HashDelete( Hash )
   TYPE(HashTable_t), POINTER :: Hash

   IF ( ASSOCIATED( Hash ) ) THEN
      IF ( ASSOCIATED( Hash % Bucket ) ) THEN
         CALL HashClean( Hash )
         DEALLOCATE( Hash % Bucket )
      END IF
      DEALLOCATE( Hash )
   END IF
 END SUBROUTINE HashDelete

!
 RECURSIVE FUNCTION HashRebuild( hash ) RESULT(Success)
   TYPE(HashTable_t), POINTER :: Hash, NewHash
   LOGICAL :: Success

   TYPE(HashEntry_t), POINTER :: entry
   INTEGER :: i

   Success = .FALSE.
   IF ( .NOT.ASSOCIATED(Hash) ) RETURN

   NewHash => HashCreate( 4*Hash % BucketSize, Hash % MaxAvgEntries )
   IF ( .NOT. ASSOCIATED( Newhash ) ) RETURN

   DO i=1,Hash % BucketSize
      Entry => Hash % Bucket(i) % Head
      DO WHILE( ASSOCIATED( Entry ) )
         IF ( .NOT. HashAdd( Newhash, Entry % Key, Entry % Value ) ) RETURN
         Entry => Entry % Next
      END DO
   END DO

   CALL HashClean( Hash )
   DEALLOCATE( Hash % Bucket )

   DEALLOCATE( Hash )
   Hash => NewHash
   Success = .TRUE.
 END FUNCTION HashRebuild

!
 FUNCTION HashValue( Hash, key ) RESULT(Value)
   TYPE(HashTable_t), POINTER :: Hash
   CHARACTER(LEN=*) :: Key
   TYPE(HashValue_t), POINTER :: Value

   INTEGER ::  n
   TYPE(HashEntry_t), POINTER :: Entry

   NULLIFY( Value )
   Entry => HashFind( Hash, key, n )
   IF ( ASSOCIATED( Entry ) ) Value => Entry % Value
 END FUNCTION HashValue

!
 SUBROUTINE HashInitWalk( Hash )
   TYPE(HashTable_t), POINTER :: Hash
   Hash % CurrentBucket = 0
   NULLIFY( Hash % CurrentEntry )
 END SUBROUTINE HashInitWalk

!
 FUNCTION HashNext( Hash ) RESULT(Entry)
   TYPE(HashTable_t), POINTER :: Hash
   TYPE(HashEntry_t), POINTER :: Entry

   LOGICAL :: Current

   NULLIFY( Entry )
   IF ( .NOT.ASSOCIATED(Hash) ) RETURN

   Current = .FALSE.
   IF ( ASSOCIATED( Hash % CurrentEntry ) ) THEN
      Current = ASSOCIATED( Hash % CurrentEntry % Next )
   END IF

   IF ( Current ) THEN
      Hash % CurrentEntry => Hash % CurrentEntry % Next
   ELSE
      Hash % CurrentBucket = Hash % CurrentBucket + 1

      DO WHILE(  Hash % CurrentBucket < Hash % BucketSize .AND. &
             .NOT.ASSOCIATED( Hash % Bucket(Hash % CurrentBucket) % Head) )
         Hash % CurrentBucket = Hash % CurrentBucket + 1
      END DO

      IF ( Hash % CurrentBucket >= Hash % BucketSize ) RETURN

      Hash % CurrentEntry => Hash % Bucket(Hash % CurrentBucket) % Head
   END IF

   Entry => Hash % CurrentEntry
 END FUNCTION HashNext

!
!void HashStats( HashTable_t *hash )
!{
!   HashEntry_t *entry;
!   int *BucketEntries;
!   int n,i,j,MaxEntries=0,MinEntries=1<<30;
!  
!   for( i=0; i<hash->BucketSize; i++ )
!   {
!      n = 0;
!      for( entry = hash->Bucket[i]; entry; entry=entry->next ) n++;
!      MaxEntries = MAX( MaxEntries,n );
!      MinEntries = MIN( MinEntries,n );
!   }
!
!   BucketEntries = (int *)calloc( MaxEntries+1,sizeof(int) );
!
!   for( i=0; i<hash->BucketSize; i++ )
!   {
!      n = 0;
!      for( entry = hash->Bucket[i]; entry; entry=entry->next ) n++;
!      BucketEntries[n]++;
!   }
!
!   fprintf( stdout, "\n\nHash table statistics:\n\n" );
!   fprintf( stdout, "Buckets: % 4d\n", hash->BucketSize );
!   fprintf( stdout, "Entries: % 4d\n", hash->TotalEntries );
!   fprintf( stdout, "Min / Bucket: %d\n", MinEntries );
!   fprintf( stdout, "Max / Bucket: %d\n\n", MaxEntries );
!
!
!   for( n=MinEntries; n<=MaxEntries; n++ )
!      if ( BucketEntries[n] )
!        fprintf( stdout, "% 4d entries in % 4d buckets.\n",
!            n,BucketEntries[n] );
!
!   free( BucketEntries );
!}
END MODULE HashTable
